@@ -38,22 +38,75 @@
 #include "tile.h"
 #include "tilelayer.h"
 
+#include <QCache>
 #include <QPaintEngine>
 #include <QPainter>
 #include <QVector2D>
 
-#include "qtcompat_p.h"
-
 #include <cmath>
 
 using namespace Tiled;
 
-static QPixmap tinted(const QPixmap &pixmap, const QColor &color)
+struct TintedKey
 {
-    if (!color.isValid() || color == QColor(255, 255, 255, 255))
+    const qint64 pixmapKey;
+    const QRect rect;
+    const QColor color;
+
+    bool operator==(const TintedKey &o) const
+    {
+        return pixmapKey == o.pixmapKey &&
+                rect == o.rect &&
+                color == o.color;
+    }
+};
+
+#if QT_VERSION < QT_VERSION_CHECK(6, 0, 0)
+uint qHash(const TintedKey &key, uint seed) Q_DECL_NOTHROW
+{
+    auto h = ::qHash(key.pixmapKey, seed);
+    h = ::qHash(key.rect.topLeft(), h);
+    h = ::qHash(key.rect.bottomRight(), h);
+    h = ::qHash(key.color.rgba(), h);
+    return h;
+}
+#else
+size_t qHash(const TintedKey &key, size_t seed) Q_DECL_NOTHROW
+{
+    return qHashMulti(seed, key.pixmapKey, key.rect, key.color.rgba());
+}
+#endif
+
+// Borrowed from qpixmapcache.cpp
+static inline qsizetype cost(const QPixmap &pixmap)
+{
+    // make sure to do a 64bit calculation; qsizetype might be smaller
+    const qint64 costKb = static_cast<qint64>(pixmap.width())
+                        * pixmap.height() * pixmap.depth() / (8 * 1024);
+    const qint64 costMax = std::numeric_limits<qsizetype>::max();
+    // a small pixmap should have at least a cost of 1(kb)
+    return static_cast<qsizetype>(qBound(1LL, costKb, costMax));
+}
+
+static bool needsTint(const QColor &color)
+{
+    return color.isValid() &&
+            color != QColor(255, 255, 255, 255);
+}
+
+static QPixmap tinted(const QPixmap &pixmap, const QRect &rect, const QColor &color)
+{
+    if (pixmap.isNull() || !needsTint(color))
         return pixmap;
 
-    QPixmap resultImage = pixmap;
+    // Cache for up to 100 MB of tinted pixmaps, since tinting is expensive
+    static QCache<TintedKey, QPixmap> cache { 100 * 1024 };
+
+    const TintedKey tintedKey { pixmap.cacheKey(), rect, color };
+    if (auto cached = cache.object(tintedKey))
+        return *cached;
+
+    QPixmap resultImage = pixmap.copy(rect);
     QPainter painter(&resultImage);
 
     QColor fullOpacity = color;
@@ -65,7 +118,7 @@ static QPixmap tinted(const QPixmap &pixmap, const QColor &color)
 
     // apply the original alpha to the final image
     painter.setCompositionMode(QPainter::CompositionMode_DestinationIn);
-    painter.drawPixmap(0, 0, pixmap);
+    painter.drawPixmap(resultImage.rect(), pixmap, rect);
 
     // apply the alpha of the tint color so that we can use it to make the image
     // transparent instead of just increasing or decreasing the tint effect
@@ -74,12 +127,19 @@ static QPixmap tinted(const QPixmap &pixmap, const QColor &color)
 
     painter.end();
 
+    cache.insert(tintedKey, new QPixmap(resultImage), cost(resultImage));
+
     return resultImage;
 }
 
 MapRenderer::~MapRenderer()
 {}
 
+QRect MapRenderer::mapBoundingRect() const
+{
+    return boundingRect(map()->tileBoundingRect());
+}
+
 QRectF MapRenderer::boundingRect(const ImageLayer *imageLayer) const
 {
     QRectF bounds = QRectF(QPointF(), imageLayer->image().size());
@@ -125,7 +185,9 @@ void MapRenderer::drawImageLayer(QPainter *painter,
                                  const QRectF &exposed) const
 {
     painter->save();
-    painter->setBrush(tinted(imageLayer->image(), imageLayer->effectiveTintColor()));
+    painter->setBrush(tinted(imageLayer->image(),
+                             imageLayer->image().rect(),
+                             imageLayer->effectiveTintColor()));
     painter->setPen(Qt::NoPen);
     if (exposed.isNull())
         painter->drawRect(boundingRect(imageLayer));
@@ -216,8 +278,8 @@ void MapRenderer::drawTileLayer(QPainter *painter, const TileLayer *layer, const
     // the grid size because this has already been taken into account by
     // boundingRect.
     QMargins drawMargins = layer->drawMargins();
-    drawMargins.setTop(drawMargins.top() - tileSize.height());
-    drawMargins.setRight(drawMargins.right() - tileSize.width());
+    drawMargins.setTop(qMax(0, drawMargins.top() - tileSize.height()));
+    drawMargins.setRight(qMax(0, drawMargins.right() - tileSize.width()));
     rect.adjust(-drawMargins.right(),
                 -drawMargins.bottom(),
                 drawMargins.left(),
@@ -228,8 +290,13 @@ void MapRenderer::drawTileLayer(QPainter *painter, const TileLayer *layer, const
     auto tileRenderFunction = [layer, &renderer, tileSize](QPoint tilePos, const QPointF &screenPos) {
         const Cell &cell = layer->cellAt(tilePos - layer->position());
         if (!cell.isEmpty()) {
-            const Tile *tile = cell.tile();
-            const QSize size = (tile && !tile->image().isNull()) ? tile->size() : tileSize;
+            QSize size = tileSize;
+
+            if (cell.tileset()->tileRenderSize() == Tileset::TileSize) {
+                if (const Tile *tile = cell.tile())
+                    size = tile->size();
+            }
+
             renderer.render(cell, screenPos, size, CellRenderer::BottomLeft);
         }
     };
@@ -239,14 +306,7 @@ void MapRenderer::drawTileLayer(QPainter *painter, const TileLayer *layer, const
 
 void MapRenderer::setFlag(RenderFlag flag, bool enabled)
 {
-#if QT_VERSION >= 0x050700
     mFlags.setFlag(flag, enabled);
-#else
-    if (enabled)
-        mFlags |= flag;
-    else
-        mFlags &= ~flag;
-#endif
 }
 
 /**
@@ -387,30 +447,42 @@ void CellRenderer::render(const Cell &cell, const QPointF &screenPos, const QSiz
         flush();
 
     const QPixmap &image = tile->image();
-    const QSizeF imageSize = image.size();
-    if (imageSize.isEmpty())
+    QRect imageRect = tile->imageRect();
+    if (imageRect.isEmpty())
         return;
 
-    const QSizeF scale(size.width() / imageSize.width(), size.height() / imageSize.height());
+    if (needsTint(mTintColor))
+        imageRect.moveTopLeft(QPoint(0, 0));
+
     const QPoint offset = tile->offset();
-    const QPointF sizeHalf = QPointF(size.width() / 2, size.height() / 2);
+    const QPointF sizeHalf { size.width() / 2, size.height() / 2 };
 
     bool flippedHorizontally = cell.flippedHorizontally();
     bool flippedVertically = cell.flippedVertically();
 
     QPainter::PixmapFragment fragment;
     // Calculate the position as if the origin is TopLeft, and correct it later.
-    fragment.x = screenPos.x() + (offset.x() * scale.width()) + sizeHalf.x();
-    fragment.y = screenPos.y() + (offset.y() * scale.height()) + sizeHalf.y();
-    fragment.sourceLeft = 0;
-    fragment.sourceTop = 0;
-    fragment.width = imageSize.width();
-    fragment.height = imageSize.height();
-    fragment.scaleX = flippedHorizontally ? -1 : 1;
-    fragment.scaleY = flippedVertically ? -1 : 1;
+    fragment.x = screenPos.x() + sizeHalf.x();
+    fragment.y = screenPos.y() + sizeHalf.y();
+    fragment.sourceLeft = imageRect.x();
+    fragment.sourceTop = imageRect.y();
+    fragment.width = imageRect.width();
+    fragment.height = imageRect.height();
+    fragment.scaleX = size.width() / imageRect.width();
+    fragment.scaleY = size.height() / imageRect.height();
     fragment.rotation = 0;
     fragment.opacity = 1;
 
+    const auto fillMode = tile->tileset()->fillMode();
+    if (fillMode == Tileset::PreserveAspectFit) {
+        const auto minScale = std::min(fragment.scaleX, fragment.scaleY);
+        fragment.scaleX = minScale;
+        fragment.scaleY = minScale;
+    }
+
+    fragment.x += offset.x() * fragment.scaleX;
+    fragment.y += offset.y() * fragment.scaleY;
+
     // Correct the position if the origin is BottomLeft.
     if (origin == BottomLeft)
         fragment.y -= size.height();
@@ -435,10 +507,16 @@ void CellRenderer::render(const Cell &cell, const QPointF &screenPos, const QSiz
         fragment.x += halfDiff;
     }
 
-    fragment.scaleX = scale.width() * (flippedHorizontally ? -1 : 1);
-    fragment.scaleY = scale.height() * (flippedVertically ? -1 : 1);
+    fragment.scaleX *= flippedHorizontally ? -1 : 1;
+    fragment.scaleY *= flippedVertically ? -1 : 1;
 
+    // Avoid using drawPixmapFragments with OpenGL in Qt 6.4.1 and above
+    // (https://bugreports.qt.io/browse/QTBUG-111416)
+#if QT_VERSION < QT_VERSION_CHECK(6, 4, 1) || QT_VERSION >= QT_VERSION_CHECK(6, 5, 0)
     if (mIsOpenGL || (fragment.scaleX > 0 && fragment.scaleY > 0)) {
+#else
+    if (!mIsOpenGL && fragment.scaleX > 0 && fragment.scaleY > 0) {
+#endif
         mTile = tile;
         mFragments.append(fragment);
         return;
@@ -457,10 +535,11 @@ void CellRenderer::render(const Cell &cell, const QPointF &screenPos, const QSiz
 
     const QRectF target(fragment.width * -0.5, fragment.height * -0.5,
                         fragment.width, fragment.height);
-    const QRectF source(0, 0, fragment.width, fragment.height);
+    const QRectF source(fragment.sourceLeft, fragment.sourceTop,
+                        fragment.width, fragment.height);
 
     mPainter->setTransform(transform);
-    mPainter->drawPixmap(target, tinted(image, mTintColor), source);
+    mPainter->drawPixmap(target, tinted(image, tile->imageRect(), mTintColor), source);
     mPainter->setTransform(oldTransform);
 
     // A bit of a hack to still draw tile collision shapes when requested
@@ -471,7 +550,7 @@ void CellRenderer::render(const Cell &cell, const QPointF &screenPos, const QSiz
         mFragments.append(fragment);
         paintTileCollisionShapes();
         mTile = nullptr;
-        mFragments.resize(0);
+        mFragments.clear();
     }
 }
 
@@ -485,7 +564,9 @@ void CellRenderer::flush()
 
     mPainter->drawPixmapFragments(mFragments.constData(),
                                   mFragments.size(),
-                                  tinted(mTile->image(), mTintColor));
+                                  tinted(mTile->image(),
+                                         mTile->imageRect(),
+                                         mTintColor));
 
     if (mRenderer->flags().testFlag(ShowTileCollisionShapes)
             && mTile->objectGroup()
@@ -494,7 +575,7 @@ void CellRenderer::flush()
     }
 
     mTile = nullptr;
-    mFragments.resize(0);
+    mFragments.clear();
 }
 
 /**
@@ -536,7 +617,7 @@ void CellRenderer::paintTileCollisionShapes()
 
     mPainter->setRenderHint(QPainter::Antialiasing);
 
-    for (const auto &fragment : qAsConst(mFragments)) {
+    for (const auto &fragment : std::as_const(mFragments)) {
         QTransform tileTransform;
         tileTransform.translate(fragment.x, fragment.y);
         tileTransform.rotate(fragment.rotation);
